import unittest
from coodict import CooDict

class TestCooDict(unittest.TestCase):

    @classmethod
    def setUp(cls):
        cls.cd = CooDict(base={1:1, 2:2, 3:'dog'})

    def test_add(self):
        """Test adding works, and ends up in overlay."""
        self.cd[4] = 'cat'
        self.assertEqual(self.cd[4], 'cat')
        self.assertIn(4, self.cd.overlay)
        self.assertNotIn(4, self.cd.base)

    def test_repr(self):
        self.cd[4] = 'cat'
        self.assertEqual(self.cd, {1:1, 2:2, 3:'dog', 4:'cat'})

    def test_keys(self):
        self.cd[4] = 'cat'
        self.assertEqual([1,2,3,4], sorted(self.cd))

    def test_len(self):
        self.cd[4] = 'cat'
        self.assertEqual(len(self.cd), 4)

    def test_clear(self):
        """Test that overlay and deleted are cleared, but not base."""
        self.cd[4] = 'cat'
        self.cd.clear()
        self.assertEqual(len(self.cd.overlay), 0)
        self.assertEqual(len(self.cd.deleted), 0)
        self.assertTrue(len(self.cd.base) > 0)

    def test_pop_no_key(self):
        """Test that popping a non-existent key raises KeyError."""
        with self.assertRaises(KeyError) as r:
            self.cd.pop(0)

    def test_pop_base_key(self):
        """Test that popping existing key sets singleton in overlay."""
        self.cd.pop(1)
        self.assertIn(1, self.cd.deleted)

    def test_pop_overlay_key(self):
        """Test that popping overlay key just removes it."""
        self.cd[4] = 'cat'
        self.cd.pop(4)
        self.assertFalse(4 in self.cd.overlay)
        self.assertFalse(4 in self.cd.deleted)

    def test_values(self):
        self.cd[4] = 'cat'
        self.assertEqual(['1','2','cat','dog'], sorted(str(x) for x in self.cd.values()))

    def test_copy(self):
        self.cd[4] = 'cat'
        copy = self.cd.copy()
        copy[5] = 'apple'
        self.assertTrue(4 in copy)
        self.assertFalse(copy is self.cd)
        self.assertNotEqual(copy, self.cd)

if __name__ == "__main__":
    unittest.main()



"""Simple speed tests for comparison of coodict to different approaches to
dict objects and 'cached/copied' access.

Compares
* dict - baseline to examine performance hits for simple operations.
* new dict - new = dict(base); new.update(overlay)
* coodict
* pickle loads/dumps - a fast (if unusual) way to copy a dict.
* copy.deepcopy - the most general (and costly) way to copy a dict.

setup generates 3 objects:
* base - contains every key and value
* overlay - contains a modified value for every 5th key
* deleted - contains a 'deletion' for every 10th key
"""

import sys
from collections import defaultdict
from timeit import timeit

number = 10000

setup = """
from coodict import CooDict
from copy import deepcopy
import pickle

base = {}
overlay = {}
deleted = set()

for i in range(0, 1000):
    base[i] = i
    if not i % 5:
        overlay[i] = i + 50000
    if not i % 10:
        deleted.add(i)

cd = CooDict(base=base, overlay=overlay)
cd.deleted = deleted

"""

def speeds():
    totals = defaultdict(int)
    def timer(code, desc="", setUp=setup, num=number):
        result = timeit(stmt=code, setup=setup, number=num)
        totals[desc] += result
        print("{:20}: {:.5f}".format(desc, result))

    print("CONTAINS")
    timer("12 in base; 12 in overlay", "dict")
    timer("new = dict(base); new.update(overlay); 12 in new", "new dict")
    timer("12 in cd", "coodict")
    timer("x=pickle.loads(pickle.dumps(base)); 12 in x", "pickle load/dump")
    timer("x=deepcopy(base); 12 in x", "deepcopy")

    print("LEN")
    timer("len(base); len(overlay)", "dict")
    timer("new = dict(base); new.update(overlay); len(new)", "new dict")
    timer("len(cd)", "coodict")
    timer("x=pickle.loads(pickle.dumps(base)); len(x)", "pickle load/dump")
    timer("x=deepcopy(base); len(x)", "deepcopy")

    print("KEYS:")
    timer("base.keys(); overlay.keys()", "dict")
    timer("new = dict(base); new.update(overlay); new.keys()", "new dict")
    timer("cd.keys()", "coodict")
    timer("x=pickle.loads(pickle.dumps(base)); x.keys()", "pickle load/dump")
    timer("x=deepcopy(base); x.keys()", "deepcopy")

    print("VALUES:")
    timer("base.values(); overlay.values()", "dict")
    timer("new = dict(base); new.update(overlay); new.values()", "new dict")
    timer("cd.values()", "coodict")
    timer("x=pickle.loads(pickle.dumps(base)); x.values()", "pickle load/dump")
    timer("x=deepcopy(base); x.values()", "deepcopy")

    print("WRITE then READ")
    timer("base[1] = 'cat'; base[1]", "dict")
    timer("new = dict(base); new.update(overlay); new[1] = 'cat'; new[1]", "new dict")
    timer("cd[1] = 'cat'; cd[1]", "coodict")
    timer("x=pickle.loads(pickle.dumps(base)); x[1] = 'cat'; x[1]", "pickle load/dump")
    timer("x=deepcopy(base); x.values(); x[1] = 'cat'; x[1]", "deepcopy")

    print("DELETED")
    timer("base[1] = 1;base.pop(1)", "dict")
    timer("new = dict(base); new.update(overlay); new[1] = 'cat'; new.pop(1)", "new dict")
    timer("cd[1] = 1;cd.pop(1)", "coodict")
    timer("x=pickle.loads(pickle.dumps(base)); x.pop(1)", "pickle load/dump")
    timer("x=deepcopy(base); x.values(); x.pop(1)", "deepcopy")

    print("TOTALS")
    for k, v in totals.items():
        print("{:20}: {:.5f}".format(k, v))


if __name__ == "__main__":
    speeds()



"""CooDict - a (Scottish) copy-on-write dictionary class."""

from collections import abc, UserDict


class CooDict(UserDict):  # pylint: disable=too-many-ancestors
    """Copy-on-write dictionary class."""

    # pylint:disable=super-init-not-called
    def __init__(self, base=None, overlay=None):
        """Set up base and data (overlay) dicts."""
        self.base = base or {}
        # alias data to overlay for clarity in code/docs
        self.data = self.overlay = overlay or {}
        self.deleted = set()

    def __delitem__(self, key):
        """Update value to Deleted class to mark deletion"""
        # Throw away keys only in overlay
        if key in self.overlay:
            self.overlay.pop(key)
        elif key in self.base and key not in self.deleted:
            self.deleted.add(key)
        else:
            raise KeyError

    def __getitem__(self, key):
        """Return item from overlay, or fall back to base."""
        if key in self.deleted:
            raise KeyError
        if key in self.overlay:
            return self.overlay[key]
        if isinstance(self.base[key], abc.Mapping):
            self.overlay[key] = self.base[key].copy()
            return self.overlay[key]
        else:
            return self.base[key]

    def __setitem__(self, key, value):
        """Clear deleted keys before setting."""
        try:
            self.deleted.remove(key)
        except KeyError:
            pass
        super().__setitem__(key, value)

    def __iter__(self):
        """CooDict iterator."""
        yield from self.base.keys() | self.overlay.keys() - self.deleted

    def __len__(self):
        """len of CooDict."""
        return len(self.base.keys() | self.overlay.keys() - self.deleted)

    def __repr__(self):
        """Representation of CooDict."""
        return repr({k: self[k] for k in self.__iter__()})

    def __str__(self):
        """String representation of CooDict."""
        return str(self.__repr__())

    def copy(self):
        """Return a copy of the CooDict (as a CooDict instance)."""
        return CooDict(base=self.base.copy(), overlay=self.overlay.copy())

    def clear(self):
        """Clear the overlay dict."""
        self.overlay.clear()
        self.deleted.clear()



#!/usr/bin/env python
"""Generate version information from git tags, archive info, or PKG-INFO.

Based on: https://github.com/Changaco/version.py

"""

from os.path import dirname, isdir, isfile, join
import re
from subprocess import CalledProcessError, check_output
import pkg_resources


def get_version():
    """Get version in any way possible."""

    prefix = ''
    tag_re = re.compile(r'\btag: %s([0-9][^,]*)\b' % prefix)
    version_re = re.compile('^Version: (.+)$', re.M)

    # Return the version if it has been injected into the file by git-archive
    version = tag_re.search('$Format:%D$')
    if version:
        return version.group(1)

    project_dir = dirname(__file__)
    pkg_file = join(project_dir, '../PKG-INFO')

    if isdir(join(project_dir, '../.git')):
        # Get the version using "git describe".
        cmd = 'git -C %s describe --tags --match %s[0-9]* --dirty' % (project_dir, prefix)
        try:
            version = check_output(cmd.split()).decode().strip()[len(prefix):]
        except CalledProcessError:
            raise RuntimeError('Unable to get version number from git tags')

        # PEP 440 compatibility
        if '-' in version:
            version = '.dev'.join(version.split('-')[:2])
        return version

    if isfile(pkg_file):
        # Extract the version from the PKG-INFO file.
        with open(pkg_file) as info_f:
            return version_re.search(info_f.read()).group(1)

    # Package was installed (os pkg, pip etc) without PKG-INFO
    return pkg_resources.get_distribution('coodict').version


if __name__ == '__main__':
    print(get_version())



import sys
from setuptools import setup, find_packages
sys.path.append("coodict")
from version import get_version

requires = []

setup(
  name = 'coodict',
  packages = find_packages(),
  install_requires=requires,
  version=get_version(),
  description = 'A (Scottish) copy-on-write dictionary wrapper class.',
  author = 'Matthew Richardson',
  author_email = 'm.richardson@ed.ac.uk',
  license = 'GNU Affero General Public License v3 or later (AGPLv3+)',
  url = 'https://github.com/mrichar1/coodict',
  keywords = ['python', 'dictionary', 'copy-on-write', 'cow'],
  classifiers = [
      'Development Status :: 5 - Production/Stable',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.6',
      'Topic :: Software Development :: Libraries :: Python Modules',
  ],
  test_suite="tests",
  )
